"""
Calculate external validation. Two partitions should be given.

"""
import math
import numpy as np


def calculate_external(partition_a, partition_b, indices=['all']):

    from collections import defaultdict

    # size of contigency table
    R = len(partition_a)
    C = len(partition_b)

    # contigency table
    contigency_table = np.zeros((R + 1, C + 1))
    # fill the contigency table
    for i in range(0, R + 1):
        for j in range(0, C):
            if i in range(0, R):
                n_common_elements = len(set(partition_a[i]).intersection(partition_b[j]))
                contigency_table[i][j] = n_common_elements
            else:
                contigency_table[i][j] = contigency_table[:, j].sum()

        contigency_table[i][j + 1] = contigency_table[i].sum()

    N = contigency_table[R][C]
    # condensed information of ct into a mismatch matrix (pairwise agreement)
    sum_all_squared = np.sum(contigency_table[0:R][:, range(0, C)] ** 2)
    sum_R_squared = np.sum(contigency_table[0:R, C] ** 2)
    sum_R = np.sum(contigency_table[0:R, C])
    sum_C_squared = np.sum(contigency_table[R, 0:C] ** 2)
    sum_C = np.sum(contigency_table[R, 0:C])
    # computing the number of pairs that are in the same cluster both in partition A and partition B
    a = 0
    for i in range(0, R):
        for j in range(0, C):
            a += contigency_table[i][j] * (contigency_table[i][j] - 1)
    a = a / 2
    # computing the number of pair in the same cluster in partition A but in different cluster in partition B
    b = ((N**2) + sum_all_squared - (sum_R_squared + sum_C_squared)) / 2
    # computing the number of pair in different cluster in partition A but in the same cluster in partition B
    c = (sum_C_squared - sum_all_squared) / 2
    # computing the number of pairs in different cluster both in partition A and partition B
    d = (sum_R_squared - sum_all_squared    ) / 2

    M = (a + b + c + d)

    indices_funcs = {'AR': adjusted_rand, 'FM': fowlkes_mallows, 'J': jaccard, 'AW': adjusted_wallace,
                     'VD': van_dongen, 'H': hubert_normalized, 'F': f_measure,
                     'VI': variation_information, 'K': Kulczynski,
                     'Phi': Phi, 'RT': Rogers_Tanimoto, 'SS': Sokal_Sneath}
    results = defaultdict()

    if isinstance(indices, str):
        indices = [x.strip() for x in indices.split(',')]

    for index in indices:
        if index == 'all' or index == 'external':
            for cvi, func in indices_funcs.items():
                if cvi in ['AR', 'AW']:
                    results[cvi] = func(a, b, c, d, sum_C, sum_R_squared, sum_C_squared, sum_all_squared, N)
                elif cvi in ['VD', 'VI']:
                    results[cvi] = func(contigency_table, R, C, N)
                else:
                    results[cvi] = func(a, b, c, d, M)
        elif index in indices_funcs.keys():
            if index in ['AR', 'AW']:
                results[index] = indices_funcs[index](a, b, c, d, sum_C,  sum_R_squared, sum_C_squared, sum_all_squared, N)
            elif index in ['VD', 'VI']:
                results[index] = indices_funcs[index](contigency_table, R, C, N)
            else:
                results[index] = indices_funcs[index](a, b, c, d, M)

    return results



def adjusted_rand(a, b , c, d, sum_C, sum_R_squared, sum_C_squared, sum_all_squared, N):
    nc = ((N*(math.pow(N, 2) + 1)) - ((N + 1)*sum_R_squared) - ((N + 1)*sum_C_squared) + (sum_all_squared / N)) / 2*(N - 1)

    adjusted_rand_index = (a + d - nc) / (a + b + c + d - nc)

    return adjusted_rand_index

def fowlkes_mallows(a, b, c, d, M):
    if ((a + b) == 0 or (a + c) == 0):
        FM = 0
    else:
        FM = a/math.sqrt(((a + b)) * ((a + c)))

    return FM

def jaccard(a, b, c, d, M):
    if (a + b + c == 0):
        jaccard = 1
    else:
        jaccard = a / (a + b + c)

    return jaccard

def adjusted_wallace(a, b, c, d, sum_C, sum_R_squared, sum_C_squared, sum_all_squared, N):
    if ((a + b) == 0):
        wallace = 0
    else:
        wallace = a / (a + b)
    SID_B = 1 - ((sum_C_squared - sum_C) / (N * (N - 1)))
    if ((SID_B) == 0):
        adjusted_wallace_index = 0
    else:
        adjusted_wallace_index = (wallace - (1 - SID_B)) / (1 - (1 - SID_B))

    return adjusted_wallace_index

def van_dongen(contigency_table, R, C, N):

    VD_i = np.sum(max(contigency_table[i, 0:C]) for i in range(R))
    VD_j = np.sum(max(contigency_table[0: R, j]) for j in range(C))

    van_dongen_index = ((2 * N) - VD_i - VD_j)/ (2*N)
    
    return van_dongen_index


def hubert_normalized(a, b, c, d, M):

    aux_hub1 = (a + b) * (a + c)
    aux_hub2 = (a + b) * (a + c) * (d + b) * (d + c)
    if aux_hub2 == 0:
        hub_normalized = float(0)
    else:
        hub_normalized = (M * a - aux_hub1) / (math.sqrt(aux_hub2))

    return hub_normalized

def f_measure(a, b, c, d, M):
    if a + b + c == 0:
        f_measure_index = float(0)
    else:
        f_measure_index = (2*a) / ((2*a) + b + c)

    return f_measure_index

def variation_information(contigency_table, R, C, N):
    mutual_info = 0
    mutual_info_aux = 0
    entropy_C = 0
    entropy_P = 0
    flag_entropy = 0
    for i in range(R):
        # calculation for entropy
        p_i = contigency_table[i][-1] / N
        if p_i != 0:
            entropy_C += (p_i * math.log10(p_i))
        for j in range(C):
            if flag_entropy == 0:
                # calculation for entropy
                p = contigency_table[-1][j] / N
                if p != 0:
                    entropy_P += (p * math.log10(p))
            p_ij = contigency_table[i][j] / N
            p_j = contigency_table[-1][j] / N
            if not 0 in [p_ij, p_i, p_j]:
                mutual_info_aux += (p_ij * math.log10(p_ij / (p_i * p_j)))
        flag_entropy = 1
        mutual_info += mutual_info_aux

    VI = -entropy_C - entropy_P - (2 * mutual_info)

    return VI


def Czekanowski_Dice(a, b, c, d, M):

    return 2*a / 2*a + b + c

def Kulczynski(a, b, c, d, M):

    c = (a / (a + c)) + (a / (a + b))

    return c/2


def Phi(a, b, c, d, M):

    c1 = (a * d) - (b * c)
    c2 = (a + b)*(a + c)*(b + d)*(c + d)

    if c2 == 0:
        c = 0
    else:
        c = c1 / c2
    return c

def Rogers_Tanimoto(a, b, c, d, M):

    return (a + d) / (a + d + 2*(b + c))


def Sokal_Sneath(a,b,c,d,M):

    return a / (a + 2*(b+c))





from sklearn.datasets import load_iris, make_blobs, load_breast_cancer, load_wine
from clusterval import Clusterval
from datasets.datasets import load_khan_train, load_khan_test, load_vote_repub, load_animals
import re

synthetic_500_k_4, _ = make_blobs(n_samples=500, centers=4, n_features=5, random_state=0)

synthetic_dim2_9 = []
pattern = re.compile(r'^\s+')
with open('/home/nuno/Documentos/IST/Tese/Datasets/data_dim_k=9_txt/dim2.txt', 'r') as dim2:
    for line in dim2:
        re_new = re.sub(pattern, '', line)
        new = ''
        for i, el in enumerate(re_new[:-1]):
            if (el != ' ') and (re_new[i+1] == ' '):
                new = new + el + ','
            elif el != ' ':
                new += el

        new = new.split(',')
        new = [float(item) for item in new]
        synthetic_dim2_9.append(new)

synthetic_dim3_9 = []
with open('/home/nuno/Documentos/IST/Tese/Datasets/data_dim_k=9_txt/dim3.txt') as s1:
    for line in s1:
        re_new = re.sub(pattern, '', line)
        new = ''
        for i, el in enumerate(re_new[:-1]):
            if (el != ' ') and (re_new[i + 1] == ' '):
                new = new + el + ','
            elif el != ' ':
                new += el

        new = new.split(',')
        new = [float(item) for item in new]
        synthetic_dim3_9.append(new)
movement_libras = []    
with open('/home/nuno/Documentos/IST/Tese/Datasets/real-world/movement_libras.data') as libra:
    for line in libra:
    	line = line.split(',')
    	movement_libras.append(line[:(len(line)-1)])

datasets = {'iris': load_iris()['data'], 'wine': load_wine()['data'], 'cancer': load_breast_cancer()['data'],
            'synthetic': synthetic_500_k_4, 'train_khan': load_khan_train()['data'], 'test_khan': load_khan_test()['data'],
            'vote.repub': load_vote_repub()['data'], 'animals': load_animals()['data'], 's_dim2_k_9': synthetic_dim2_9,
            's_dim3_k_9': synthetic_dim3_9, 'libras':movement_libras}

linkage = ['ward', 'single', 'centroid', 'average', 'complete']


c = Clusterval(min_k=2, max_k=8, index='internal')

for name, dataset in datasets.items():
    if name == 'iris':
        print('Testing for {0}\n'.format(name))
        for l in linkage:
            print('{0} with linkage {1}\n'.format(name, l))
            c.link = l
            eval = c.evaluate(dataset)
            print(eval.long_info)
        print('\n\n')



"""
Function evaluate does hierarchical clustering and validation of it's results.
Validation can be done with external or internal indices, or both.


"""


from scipy.cluster.hierarchy import dendrogram, fcluster, cut_tree
from fastcluster import linkage
from scipy.spatial.distance import pdist
from sklearn.cluster import KMeans
from sklearn.neighbors import NearestCentroid
from statistics import mean
import pandas as pd
from collections import defaultdict
import matplotlib.pyplot as plt
import itertools
import numpy as np
import math
import random

from clusterval.internal import calculate_internal
from clusterval.external import calculate_external


class Clusterval:
    """
    min_k: int, optional
    minimum number of clusters to test. Default 2.

    max_k: int, optional
    maximum number of clusters to test. Default 8.

    algorithm: str, optional
    clustering algorithm to use. Accepts 'single', 'ward'(default), 'complete', 'centroid', 'average' and 'kmeans

    bootstrap_samples: int, optional
    number of bootstrap samples simulated. Default 250

    index: list of str or str, optional
    what indices to be calculated. Accepts 'all' to calculate all, 'internal' to calculate only internal indices,
     'external' to calculate external indices, or one of ['AR', 'FM', 'J', 'AW', 'VD', 'H', 'F', 'VI', 'K', 'Phi', 'RT', 'SS',
     'CVNN', 'XB', 'S_Dbw', 'DB', 'S', 'SD', 'PBM', 'Dunn']. Default 'all'.
    """

    def __init__(self, min_k=2, max_k=8, algorithm='ward', bootstrap_samples=250, index='all'):

        external_indices = ['AR', 'FM', 'J', 'AW', 'VD', 'H', 'F', 'VI', 'K', 'Phi', 'RT', 'SS']

        internal_indices = ['CVNN', 'XB', 'S_Dbw', 'DB', 'S', 'SD', 'PBM', 'Dunn']

        min_indices = ['VD', 'VI', 'MS', 'CVNN', 'XB', 'S_Dbw', 'DB', 'SD']

        idx_distance_matrix = ['XB', 'S_Dbw', 'DB', 'SD', 'PBM']

        indices = {'AR': ['AR'], 'FM': ['FM'], 'J': ['J'], 'AW': ['AW'], 'VD': ['VD'], 'H': ['H'],
                   'F': ['F'], 'K': ['K'], 'Phi': ['Phi'], 'RT': ['RT'], 'SS': ['SS'],
                   'VI': ['VI'], 'CVNN': ['CVNN'], 'XB': ['XB'],
                   'S_Dbw': ['S_Dbw'], 'DB': ['DB'], 'S': ['S'], 'SD': ['SD'],'PBM': ['PBM'], 'Dunn': ['Dunn'],
                   'all': external_indices + internal_indices, 'external': external_indices,
                   'internal': internal_indices
                   }


        if isinstance(index, str):
            index = [x.strip() for x in index.split(',')]

        for idx in index:
            if idx not in indices.keys():
                raise ValueError('{0} is not a valid index value, please check help(clusterval.Clusterval) to see acceptables indices'.format(
                    idx))

        self._external_indices = external_indices
        self._internal_indices = internal_indices
        self._min_indices = min_indices
        self._indices = indices
        self._idx_distance_matrix = idx_distance_matrix

        self.min_k = min_k
        self.max_k = max_k
        self.algorithm = algorithm
        self.bootstrap_samples = bootstrap_samples
        self.index = index

        columns = []
        for i in self.index:
            columns = columns + self._indices[i]
        self.output_df = pd.DataFrame(
            index=range(self.min_k, self.max_k + 1),
            columns=columns,
            dtype=np.float64
        )
        self.final_k = None
        self.count_choice = None
        self.final_clusters = None
        self.Z= None

        self.long_info = None

    def __repr__(self):
        args = ['{}={}'.format(' ' + key, value) for key,value in self.__dict__.items() if key not in ['output_df',
                                                                                                       '_external_indices',
                                                                                                       '_internal_indices',
                                                                                                       '_min_indices',
                                                                                                       '_indices',
                                                                                                       '_data',
                                                                                                       'final_k',
                                                                                                       'count_choice',
                                                                                                       'final_clusters',
                                                                                                       'long_info',
                                                                                                       'Z',
                                                                                                       '_idx_distance_matrix']]
        args = ','.join(args)
        return 'Clusterval(' + str(args) + ')\nfinal_k = {}'.format(self.final_k)

    def __str__(self):
        args = ['{} is {}'.format(' ' + key, value) for key, value in self.__dict__.items() if key not in ['output_df',
                                                                                                       '_external_indices',
                                                                                                       '_internal_indices',
                                                                                                       '_min_indices',
                                                                                                       '_indices',
                                                                                                       '_data',
                                                                                                       'final_k',
                                                                                                       'count_choice',
                                                                                                       'final_clusters',
                                                                                                        'long_info',
                                                                                                           'Z',
                                                                                                           '_idx_distance_matrix']]
        args = ';\n'.join(args)
        return 'Clusterval: \n' + str(args) + '\n'


    def _cluster_indices(self, cluster_assignments, idx):
        '''
        Transform cluster memebership array into array of clusters
        :param cluster_assignments: array
        :param idx: array with indices
        :return: array of clusters
        '''

        n = int(cluster_assignments.max())
        clusters = []
        for cluster_number in range(0, n + 1):
            aux = np.where(cluster_assignments == cluster_number)[0].tolist()
            if aux:
                cluster = list(idx[i] for i in aux)
                clusters.append(cluster)
        return clusters

    def _distance_dict(self, data):
        """
        Build dictionary that maps pairs to distance values
        :param data: ndarray m * (m - 1)) // 2 like returned by pdist
        :return: dictionary of tuples
        """
        from itertools import combinations
        from sympy import Symbol, solve

        diss_matrix = False
        #check if distances already calculated
        if data.ndim == 1:
            pairwise_distances = [i for i in data]

            x = Symbol('x')
            n = solve(x ** 2 - x - 2*data.shape[0])[1]

            diss_matrix = True

        else:
            pairwise_distances = pdist(data)
            n = len(data)

        comb = list(combinations([i for i in range(0, n)], 2))
        dist_dict = defaultdict(float)

        for pair, dist in zip(comb,pairwise_distances):
            dist_dict[pair] = dist

        return dist_dict, diss_matrix

    def _choose(self, choices_dict):

        for metrics in self.index:
            for metric in self._indices[metrics]:
                if metric in self.output_df.columns.values:
                    if metric in self._min_indices:
                        value = self.output_df.loc[self.output_df[metric].idxmin()]

                    else:
                        value = self.output_df.loc[self.output_df[metric].idxmax()]
                    choices_dict[metric] = [value.name, value[metric]]

        return choices_dict




    def evaluate(self, data):
        """
        Perform hierarchical clustering or Kmeans clustering on the data and calculate the validation indices
        :param data: array-like, n_samples x n_features or 1-d dissimilarity array
        Dataset to cluster
        :return: Clusterval object
        """

        columns = []
        for i in self.index:
            columns = columns + self._indices[i]
        self.output_df = pd.DataFrame(
            index=range(self.min_k, self.max_k + 1),
            columns=columns,
            dtype=np.float64
        )
        #convert data to numpy array
        data = np.array(data)
        #build dictionary with pairwise distances
        distances, diss_matrix = self._distance_dict(data)

        for metrics in self.index:
            for idx in self._indices[metrics]:
                if (diss_matrix) and (idx in self._idx_distance_matrix):
                    self.output_df = self.output_df.drop([idx], axis=1)

        clustering = defaultdict(dict)
        # dictionary with all mean values of the metrics for every k
        choices_dict = defaultdict(list)
        self.count_choice = defaultdict(int)
        results = {k: {} for k in range(self.min_k, self.max_k + 1)}

        # build dictionaries that hold the calculations

        for k in range(self.min_k, self.max_k + 1):
            for metrics in self.index:
                for idx in self._indices[metrics]:
                    if (diss_matrix) and (idx in self._idx_distance_matrix):
                        continue

                    else:
                        results[k][idx] = []
        if not any(self.output_df.columns.values):
            raise  ValueError('No CVIs being evaluated. If inputing distance matrix, please be aware that SD, XB, S_Dbw, PBM and DB are not possible to calculate.')

        if self.algorithm in ['ward', 'single', 'complete', 'average', 'centroid']:
            self.Z = linkage(data, method=self.algorithm)


        for k in range(self.min_k, self.max_k + 1):

            if self.algorithm in ['ward', 'single', 'complete', 'average', 'centroid']:
                # builds a list of the hierarchical clusters

                # with cut_tree
                partition = cut_tree(self.Z, n_clusters=k)
                clusters = self._cluster_indices(partition, [i for i in range(0, len(data))])
                clf = NearestCentroid()

                if not diss_matrix:
                    try:
                        clf.fit(data, list(itertools.chain.from_iterable(partition)))

                    #if only one cluster formed, try with fcluster - this happened with centroid
                    except ValueError as e:
                        if isinstance(e, ValueError):
                            partition = fcluster(self.Z, t=k, criterion='maxclust')
                            clusters = self._cluster_indices(partition, [i for i in range(0, len(data))])
                            clf.fit(data, list(itertools.chain(partition)))

                    centroids = clf.centroids_
                else:
                    centroids = None



            elif self.algorithm == 'kmeans':

                if diss_matrix:
                    raise ValueError('k-means does not work with only pairwise distances as input. Please set another algorithm.')

                partition = KMeans(n_clusters=k, random_state=0).fit(data)
                clusters = self._cluster_indices(partition.labels_, [i for i in range(0, len(data))])
                centroids = partition.cluster_centers_

            else:
                raise ValueError(self.algorithm + ' is not an acceptable clustering algorithm, please choose \'single\','
                                                  ' \'ward\', \'centroid\', \'complete\', \'average\' or \'kmeans\'')


            # dictionary of clustering of each 'k', to be used in internal validation
            clustering[k] = {'clusters': clusters, 'centroids': centroids}
            if (self.index[0] in ['all', 'external']) or (set(self.index).intersection(self._external_indices)):

                # external validation step
                for i in range(self.bootstrap_samples):
                    sample = random.sample(list(data), int(3 / 4 * len(data)))
                    #check if combinations are good in case of data == distance matrix
                    if diss_matrix:
                        NN = len(sample)
                        N = int(math.ceil(math.sqrt(NN*2)))
                        comb = (N*(N-1)//2)
                        if comb != NN:
                            sample = sample[:-(abs(comb - NN))]

                    if self.algorithm in ['ward', 'single', 'complete', 'average', 'centroid']:

                        Z_sample = linkage(sample, method=self.algorithm)

                        #clusters_sample = self._cluster_indices(fcluster(Z_sample, t=k, criterion='maxclust'), [i for i in range(0, len(sample))])

                        clusters_sample = self._cluster_indices(cut_tree(Z_sample, k), [i for i in range(0, len(sample))])

                    elif diss_matrix:
                        raise ValueError('k-means does not work with only pairwise distances as input. Please set another algorithm.')

                    else:
                        clusters_sample = self._cluster_indices(KMeans(n_clusters=k, random_state=0).fit(sample).labels_, [i for i in range(0, len(sample))])

                    external = calculate_external(clusters, clusters_sample, indices=self.index)

                    for key, metric in external.items():
                        if (key in self.index) or (self.index):
                            results[k][key].append(metric)

        if (self.index[0] in ['all', 'internal']) or (set(self.index).intersection(self._internal_indices)):
            # dictionary of distance between pairs of data points, e.g, {(pair1,pair2) : dist(pair1,pair2)}

            # internal validation step
            if diss_matrix:
                internal = calculate_internal(data, distances, clustering, indices=['CVNN','S','Dunn'])
            else:
                internal = calculate_internal(data, distances, clustering, indices=self.index)

            for cvi, k_clust in internal.items():
                for n_c, val in k_clust.items():
                    results[n_c][cvi] = val

        for k, keys in results.items():
            for key, value in keys.items():
                if key in self._external_indices:
                    self.output_df.loc[k, key] = mean(value)

                else:
                    self.output_df.loc[k, key] = value

        choices_dict = self._choose(choices_dict)
        for values in choices_dict.values():
            self.count_choice[str(values[0])] += 1

        max_value = 0
        for key, value in self.count_choice.items():
            if value > max_value:
                max_value = value
                final_k = key
        self.final_k = int(final_k)

        if self.algorithm in ['ward', 'single', 'complete', 'average', 'centroid']:
            #self.final_clusters = fcluster(self.Z, t=self.final_k, criterion='maxclust')
            self.final_clusters = np.concatenate(np.asarray(cut_tree(self.Z, self.final_k)))
        else:
            self.final_clusters = KMeans(n_clusters=self.final_k, random_state=0).fit(data).labels_
        self.long_info = self.print_results()

        return self

    def print_results(self):
        '''
        print detailed information on the clustering results conclusion
        :return: str with the results
        '''

        output_str = '* Minimum number of clusters to test: ' + str(self.min_k) + '\n'
        output_str += '* Maximum number of clusters to test: ' + str(self.max_k) + '\n'
        output_str +='* Number of bootstrap samples generated: ' + str(self.bootstrap_samples) + '\n'
        output_str += '* Clustering algorithm used: ' + str(self.algorithm) + '\n'

        if self.index not in ['all', 'internal', 'external']:
            idx = list(self.output_df.columns.values)
        else:
            idx = self.index
        output_str +='\n* Validation Indices calculated: ' + str(idx) + '\n\n'


        output_str +="* Among all indices: \n\n"

        output_str +='\n\n* According to the majority rule, the best number of clusters is ' + str(self.final_k) + '\n\n\n'

        for k_num in sorted(self.count_choice):
            output_str +='* ' + str(self.count_choice[k_num]) + ' proposed ' + k_num + ' as the best number of clusters \n\n'

        output_str +='\t\t\t***** Conclusion *****\t\t\t\n'

        #Display all dataframe
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)
        pd.set_option('display.width', None)
        pd.set_option('display.max_colwidth', None)

        output_str +=str(self.output_df) + '\n'

        output_str +='\n* The best partition is:\n'
        output_str +=str(self.final_clusters) + '\n'

        return output_str

    def _calculate_dendrogram(self, labels=None):
        '''
        Calculate dendrogram for object linkage product
        :param labels: array-like, shape 1 x len(data.shape[0])
        :return: dendrogram
        '''

        dend = dendrogram(

            self.Z,
            # truncate_mode = 'lastp',
            # p=6,
            leaf_rotation=90.,  # rotates the x axis labels
            leaf_font_size=10,  # font size for the x axis labels
            labels=labels
        )

        return dend

    def plot(self, labels=None):
        '''
        print the hierarchical clustering dendrogram using matplotlib
        :return: matplotlib image of dendrogram
        '''

        fig = plt.figure(figsize=(40, 20))
        plt.title('Hierarchical Clustering Dendrogram - Linkage: %s, Metrics: %s' % (self.algorithm, str(self.index)),
                  fontsize=30)
        plt.xlabel('data point index', labelpad=20, fontsize=30)
        plt.ylabel('distance', labelpad=10, fontsize=30)
        plt.xticks(size=40)
        plt.yticks(size=40)
        self._calculate_dendrogram(labels)

        plt.show()




"""
Calculate internal validation. distance_dictset with pairs and their distances should be given.

"""
import itertools
from collections import defaultdict
import math
import clusterval
from scipy.spatial.distance import euclidean
import numpy as np

def calculate_internal(data, distance_dict, clustering, indices=['all']):
    """
    :param data: dataset being evaluated
    :param distance_dict: dictionary with distance between pairs
    :param clustering: dictionary with clustering results and centroids for a range of k number of clusters
    :param index: str, which index to calculate
    :return: dictionary with indices values for a range of k number of clusters
    """

    indices_funcs = {'CVNN': cvnn, 'XB': xb_improved, 'S_Dbw': s_dbw, 'DB': db_improved, 'S': silhouette, 'SD': sd,
                     'PBM': pbm, 'Dunn': dunn}
    results = defaultdict(dict)


    if isinstance(indices, str):
        indices = [x.strip() for x in indices.split(',')]

    for index in indices:
        if index == 'all' or index == 'internal':
            for cvi, func in indices_funcs.items():
                results[cvi] = func(clustering, data, distance_dict)
        elif index in indices_funcs.keys():
            results[index] = indices_funcs[index](clustering, data, distance_dict)


    return results


def cvnn(clustering, data, distance_dict):

    """

    Metric based on the calculation of intercluster separaration and intracluster compatness.
    Lower value of this metric indicates a better clustering result.

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: CVNN index
    """
    comp = defaultdict(float)
    sep = defaultdict(float)


    #nearest neighbours values to test
    nn_values = [1, 5, 10, 15, 20]

    #calculate nearest neighbours for k=[1,5,10,15,20]
    nn_history = getknn(distance_dict, nn_values)

    for k, clusters in clustering.items():

        compactness = cvnn_compactness(distance_dict, clusters['clusters'])
        separation = cvnn_separation(clusters['clusters'], nn_history, nn_values)

        comp[k] = compactness
        sep[k] = separation


    maxcomp = max(i for i in comp.values())
    maxsep = max(i for i in sep.values())


    if maxcomp == 0 or maxsep == 0:
        maxcomp = 1
        maxsep = 1

    comp_final = {key: value / maxcomp for key, value in comp.items()}
    sep_final = {key: value / maxsep for key, value in sep.items()}

    cvnn_index = {key: sep_final[key] + comp_final[key] for key in comp.keys()}


    return cvnn_index


def getknn(distance_dict, nn_values):

    nn_history = {'1': defaultdict(list), '5': defaultdict(list), '10': defaultdict(list), '15': defaultdict(list), '20': defaultdict(list)}

    #get first row index of distance_dict
    first_el = next(iter(distance_dict))[0]
    elements = [first_el]
    for key in distance_dict.keys():
        if key[0] != first_el:
            break
        else:
            elements.append(key[1])

    for k in nn_values:
        for el in elements:
            nn = []
            pairs = {key : value for key, value in distance_dict.items() if el in key}
            for k_aux in range(0, k):
                try:
                    key_min = min(pairs.keys(), key=(lambda x: pairs[x]))
                    if key_min[0] == el:
                        nn.append(key_min[1])
                    else:
                        nn.append(key_min[0])
                    del pairs[key_min]
                except:
                    raise ValueError("There is empty clusters being generated, please set max_k to a lower values (default=8)")
            nn_history[str(k)][str(el)] = nn

    return nn_history


def cvnn_compactness(distance_dict, clusters):
    compactness = 0
    sum_of_objects = 0
    for cluster in clusters:
        n_i = len(cluster)
        if n_i == 0:
            raise ValueError("Empty clusters being formed")
        sum_of_objects += n_i * (n_i - 1)
        # intracluster compactness
        pairs = list(itertools.combinations(cluster, 2))
        distance_i = pairwise_distance(pairs, distance_dict)
        compactness += distance_i


    return compactness / sum_of_objects


def cvnn_separation(clusters, nn_history, nn_values):
    best_k_value = []
    for k in nn_values:
        separation = []
        for cluster in clusters:
            n_i = len(cluster)
            if n_i == 0:
                raise ValueError("Empty clusters being formed")
            # intercluster separation
            sum_of_weights = 0
            for el in cluster:
                count_nn = 0
                for n in nn_history[str(k)][str(el)]:
                    if n not in cluster:
                        count_nn += 1
                sum_of_weights += count_nn / k

            separation.append(sum_of_weights / n_i)

        best_k_value.append(max(separation))

    return min(best_k_value)

def pairwise_distance(pairs, distance_dict):
    sum_of_distances = 0
    for pair in pairs:
        if pair not in distance_dict.keys():
            pair = (pair[1], pair[0])
        sum_of_distances += float(distance_dict[pair])
    return sum_of_distances

def xb_improved(clustering, data, distance_dict): #BIB: New indices for cluster validity assessment - Kim
    '''
    The Xie-Beni improved index (XB) defines the intercluster separation as the minimum square distance between
    cluster centers, and the intracluster compactness as the maximum square distance between each distance_dict object and its
    cluster center. The optimal cluster number is reached when the minimum of XB is found

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: Xie Beni improved index.
    '''

    xb_improved_index = defaultdict(float)
    for k, clusters in clustering.items():
        max_dist_elements = []
        min_dist_clusters = []
        n_c = len(clusters['clusters'])
        for i, cluster in enumerate(clusters['clusters']):
            sum_distances_k = 0
            n_k = len(cluster)
            if n_k == 0:
                raise ValueError("empty clusters being formed, please set a lower value for max_k")
            for el in cluster:
                dist = euclidean(data[el], clusters['centroids'][i])
                dist2 = math.sqrt(math.pow(dist, 2))
                sum_distances_k += dist2

            max_dist_elements.append(sum_distances_k / n_k)

            if i == (n_c - 1):
                continue
            else:
                for j in range(i + 1, n_c):
                    dist_centroids = euclidean(clusters['centroids'][i], clusters['centroids'][j])
                    min_dist_clusters.append(math.sqrt(math.pow(dist_centroids, 2)))


        xb_improved_index[k] = max(max_dist_elements) / min(min_dist_clusters)

    return xb_improved_index



def s_dbw(clustering, data, distance_dict):
    '''
    The S Dbw index (S Dbw) takes density into account to measure the intercluster separation.
    The basic idea is that for each pair of cluster centers, at least one of their densities should be larger
    than the density of their midpoint. The intracluster compactness is based on variances of cluster objects
    The index is the summation of these two terms and the minimum value of S Dbw indicates the
    optimal cluster number.


    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: S_Dbw index.
    '''

    s_dbw_index = defaultdict(float)

    centroid_dataset = np.mean(data, axis=0)
    variance_d = calculate_variance(data, [i for i in range(len(data))], centroid_dataset)

    for k, clusters in clustering.items():
        s_dbw_index[k] = scat(data, clusters['clusters'], clusters['centroids'], variance_d)\
                         + dens_bw(clusters['clusters'], data, clusters['centroids'])

    return s_dbw_index

def calculate_variance(data, cluster, mean):
    dist = sum(euclidean(data[el], mean) for el in cluster)
    variance = dist / len(cluster)

    return math.sqrt(variance)

def scat(data, clusters, centroids, variance_d):
    #scat improved from BIB: New indices for cluster validity assessment - Kim

    variance_clusters = []
    for ci, cluster in enumerate(clusters):
        variance_ci = calculate_variance(data, cluster, centroids[ci])
        variance_clusters.append(variance_ci/variance_d)

    return max(variance_clusters)


def dens_bw(clusters, data, centroids):

    n_c = len(clusters)
    avg_std_deviation = avg_stdev(data, clusters, centroids)
    result_sum = 0.0

    for i, cluster in enumerate(clusters):
        if i == (n_c - 1):
            break
        s = 0.0
        for j in range(i+1, n_c):
            u_ij = [(c_i + c_j)/2 for c_i,c_j in zip(centroids[i],centroids[j])]
            n_ij = cluster + clusters[j]
            dens_ij = density(data,n_ij, u_ij, avg_std_deviation)
            dens_i = density(data,cluster, centroids[i], avg_std_deviation)
            dens_j = density(data,clusters[j], centroids[j], avg_std_deviation)

            if max(dens_i, dens_j) == 0:
                #penalty for clusters with objects very far from center
                res = 10.0
            else:
                res = dens_ij / max(dens_i, dens_j)
            s += res

        result_sum += s

    return result_sum / (n_c * (n_c - 1))


def density(data, cluster, center, avg_stdev):

    sum_density = 0

    for el in cluster:
        if euclidean(data[el], center) <= avg_stdev:
            sum_density += 1

    return sum_density

def avg_stdev(data, clusters, centroids):

    average_stdev = 0.0
    for i, cluster in enumerate(clusters):
        dist = sum(euclidean(data[el], centroids[i]) for el in cluster)
        stdev = dist / len(cluster)
        #stdev = math.sqrt((dist / len(cluster)))
        average_stdev += math.sqrt(stdev)

    return average_stdev / len(clusters)

def db_improved(clustering, data, distance_dict):
    '''
    This index is obtained by averaging all cluster similarities. A smaller value indicates a better clustering.

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: Davies-Bouldin improved index.
    '''

    db_improved_index = defaultdict(float)
    for k, clusters in clustering.items():
        n_c = len(clusters['clusters'])

        dict_sim = defaultdict(float)

        sum_of_similarities = 0
        for i, cluster in enumerate(clusters['clusters']):
            if i == (n_c - 1):
                break
            rij = []
            distance_centroids = []
            if dict_sim[str(i)]:
                s_i = dict_sim[str(i)]
            elif len(cluster) == 0:
                raise ValueError("Empty clusters being formed. Set a lower value for max_k")

            elif len(cluster) == 1:
                s_i = 0
                dict_sim[str(i)] = s_i
            else:
                s_i = similarity(cluster, data, clusters['centroids'][i])
                dict_sim[str(i)] = s_i

            for j in range(i+1,n_c):
                if dict_sim[str(j)]:
                    s_j = dict_sim[str(j)]
                elif len(clusters['clusters'][j]) == 0:
                    raise ValueError("Empty clusters being formed. Set a lower value for max_k")

                elif len(clusters['clusters'][j]) == 1:
                    s_j = 0
                    dict_sim[str(j)] = s_j
                else:
                    s_j = similarity(clusters['clusters'][j], data, clusters['centroids'][j])
                    dict_sim[str(j)] = s_j

                dissimilarity = euclidean(clusters['centroids'][i], clusters['centroids'][j])

                rij.append((s_i + s_j)/dissimilarity)


            sum_of_similarities += max(rij)

        db_improved_index[k] = sum_of_similarities / n_c

    return db_improved_index

def similarity(cluster, data, centroid):

    result = sum(euclidean(data[el], centroid) for el in cluster)
    sim = result / len(cluster)
    
    return sim

def silhouette(clustering, data, distance_dict):
    '''
     Validates the clustering performance based on the pairwise difference of between and within cluster distances.
     The optimal cluster number is determined by maximizing the value of this index.

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: Silhouette index.
    '''

    silhouette_index = defaultdict(float)
    for k, clusters in clustering.items():
        n_c = len(clusters['clusters'])
        sum_clusters_diff = 0
        for i, cluster in enumerate(clusters['clusters']):
            n_i = len(cluster)
            if n_i == 0:
                raise ValueError("Empty clusters being formed. Set a lower value for max_k")

            sum_pairwise = 0
            for el in cluster:
                b = silhouette_b([x for y, x in enumerate(clusters['clusters']) if i != y], el, distance_dict)
                a = silhouette_a([x for x in cluster if x != el], el, distance_dict)

                if max(a,b) == 0:
                    sum_pairwise += 0
                else:
                    sum_pairwise += ((b - a) / max(a,b))
            sum_clusters_diff += sum_pairwise / n_i

        silhouette_index[k] = sum_clusters_diff / n_c

    return silhouette_index


def silhouette_a(cluster, el, distance_dict):
    n_i = len(cluster)
    sum_dist = 0
    for c_i in cluster:
        if c_i == el:
            continue
        pair = (el, c_i)
        if pair not in distance_dict.keys():
            pair = (pair[1], pair[0])
        sum_dist += float(distance_dict[pair])
    if n_i <= 1:
        result = sum_dist
    else:
        result = sum_dist / n_i

    return result

def silhouette_b(clusters, el, distance_dict):
    array_of_between_clusters = []
    if len(clusters) == 0:
        raise ValueError("Empty clusters being formed. Set a lower value for max_k")
    for cluster in clusters:
        sum_dist_within = 0
        n_j = len(cluster)
        if n_j == 0:
            raise ValueError("Empty clusters being formed. Set a lower value for max_k")
        for c_j in cluster:
            pair = (el, c_j)
            if pair not in distance_dict.keys():
                pair = (pair[1], pair[0])
            sum_dist_within += float(distance_dict[pair])

        array_of_between_clusters.append(sum_dist_within / n_j)

    return min(array_of_between_clusters)


def sd(clustering, data, distance_dict):
    '''
    Based on the concepts of average scattering, which indicates the compactness between clusters
    and the total separation of clusters, which indicates the separation between the items of a cluster.
    The optimal cluster number is determined by minimizing the value of this index.

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: SD index.
    '''

    sd_index = defaultdict(float)

    centroid_dataset = np.mean(data, axis=0)
    variance_d = calculate_variance(data, [i for i in range(len(data))], centroid_dataset)

    # Maximum number of input clusters.
    c_max = clustering[next(reversed(clustering))]['centroids']

    for k, clusters in clustering.items():

        sd_index[k] = dis(clusters['centroids']) + scat(data, clusters['clusters'], clusters['centroids'], variance_d)*dis(c_max)

    return sd_index


def dis(centers):
    '''
    Total separation between clusters.
    :param centers: list
    Clusters centroids
    :return: float
    '''

    sum_total_dist = 0.0
    d_max = 0.0
    d_min = math.inf
    for i, center in enumerate(centers):
        if i == (len(centers) - 1):
            break
        sum_dist = 0.0
        for j in range(i+1,len(centers)):
            dist = euclidean(center, centers[j])
            if dist > d_max:
                d_max = dist
            if dist < d_min:
                d_min = dist
            sum_dist += dist

        sum_total_dist += (1/sum_dist)

    return (d_max / d_min) * sum_total_dist


def pbm(clustering, data, distance_dict):
    '''
    The PBM index (acronym constituted of the initals of the names of its authors,
    Pakhira, Bandyopadhyay and Maulik) is calculated using the distances between
    the points and their cluster centers and the distances between the cluster centers
    themselves.

    :param clustering: dictionary with clustering results and respective centroids for each k simulation.
    :param data: dataset being analysed
    :param distance_dict: dictionary with distance between pairs
    :return: PBM index.
    '''

    pbm_index = defaultdict(float)

    centroid_dataset = np.mean(data, axis=0)
    all_elements = [i for i in range(len(data))]
    e_t = sum(euclidean(data[el],centroid_dataset) for el in all_elements)

    for k, clusters in clustering.items():
        sum_total_dist = 0.0
        d_max = 0.0
        e_c = 0.0
        for i, cluster in enumerate(clusters['clusters']):
            e_c += sum(euclidean(data[el], clusters['centroids'][i]) for el in cluster)
            if i == (len(clusters['clusters']) - 1):
                break
            for j in range(i + 1, len(clusters['clusters'])):
                dist = euclidean(clusters['centroids'][i], clusters['centroids'][j])
                if dist > d_max:
                    d_max = dist

        pbm_index[k] = math.pow((1/k) * (e_t/e_c) * d_max, 2)

    return pbm_index


def dunn(clustering, data, distance_dict):
    '''
        Dunn's calculates the minimum distance between clusters to measure the intercluster separation and the maximum
        diameter among all clusters to measure the intracluster compactness.

        :param clustering: dictionary with clustering results and respective centroids for each k simulation.
        :param data: dataset being analysed
        :param distance_dict: dictionary with distance between pairs
        :return: Dunn index.
        '''

    dunn_index = defaultdict(float)

    for k, clusters in clustering.items():

        #get maximum diameter among all clusters
        diameter_all = []
        for c in clusters['clusters']:
            if len(c) >= 2:
                pairs = list(itertools.combinations(c, 2))
                distances = []
                for pair in pairs:
                    if pair not in distance_dict.keys():
                        pair = (pair[1], pair[0])
                    distances.append(float(distance_dict[pair]))

                diameter_all.append(max(distances))

        max_diameter = max(diameter_all)

        compare_clusters = []
        for i, cluster in enumerate(clusters['clusters']):
            if i == (k - 1):
                break
            dist_between_clusters = []
            for j in range(i+1, k):
                elements = list(itertools.product(cluster, clusters['clusters'][j]))
                min_dist = math.inf
                for pair in elements:
                    if pair not in distance_dict.keys():
                        pair = (pair[1], pair[0])
                    dist = float(distance_dict[pair])
                    if dist < min_dist:
                        min_dist = dist
                dist_between_clusters.append(min_dist / max_diameter)

            compare_clusters.append(min(dist_between_clusters))

        dunn_index[k] = min(compare_clusters)

    return dunn_index



# -*- coding: utf-8 -*-



from setuptools import setup, find_packages



install_requires = [
    'sklearn', 'scipy', 'numpy', 'pandas', 'matplotlib','sympy', 'fastcluster'
]

with open('README.md') as f:
    long_description = f.read()


setup(
    name='clusterval',
    version='1.0.4',
    description='Package useful for clustering validation',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Nuno Silva',
    author_email='nuno.da.silva@tecnico.ulisboa.pt',
    url='https://github.com/Nuno09/clusterval',
    license='LICENSE',
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=install_requires,
    tests_require='pytest',
    setup_requires='pytest-runner',
    package_data={'clusterval': ['datasets/*.csv']}
)




'''
Module that imports datasets to be used in clustering.
Available datasets:
- khan - Microarray gene expression data set from Khan et al., 2001. Subset of 306 genes. Includes training and test set.
- vote.repub - Votes for Republican Candidate in Presidential Elections in each of the 50 states (1856 - 1976)
- animals - Attributes of 20 Animals

'''

import csv
import os
import numpy as np


def load_khan_train():

    train_khan = {'labels': [], 'data': []}

    with open(os.path.join(os.path.dirname(__file__), "train_khan.csv"), 'r') as file1:
        reader1 = csv.reader(file1)
        for line in reader1:
            train_khan['labels'].append(line[0])
            train_khan['data'].append(line[1:])

    train_khan['data'] = np.array(train_khan['data'])

    return train_khan



def load_khan_test():

    test_khan = {'labels': [], 'data': []}

    with open(os.path.join(os.path.dirname(__file__), "test_khan.csv"), 'r') as file2:
        reader2 = csv.reader(file2)
        for line in reader2:
            test_khan['labels'].append(line[0])
            test_khan['data'].append(line[1:])

    test_khan['data'] = np.array(test_khan['data'])

    return test_khan



def load_vote_repub():

    vote_repub = {'labels': [], 'data': []}

    with open(os.path.join(os.path.dirname(__file__), "vote_repub.csv"), 'r') as file3:
        reader3 = csv.reader(file3)
        next(reader3)
        for line in reader3:
            if 'NA' in line:
                for i, el in enumerate(line):
                    if el == 'NA':
                        line[i] = str(float(0))
            vote_repub['labels'].append(line[0])
            vote_repub['data'].append(line[1:])

    vote_repub['data'] = np.array(vote_repub['data'])

    return vote_repub


def load_animals():

    animals = {'labels': [], 'data': []}

    with open(os.path.join(os.path.dirname(__file__), "animals.csv"), 'r') as file4:
        reader4 = csv.reader(file4)
        next(reader4)
        for line in reader4:
            if 'NA' in line:
                for i, el in enumerate(line):
                    if el == 'NA':
                        line[i] = '0'
            animals['labels'].append(line[0])
            animals['data'].append(line[1:])

    animals['data'] = np.array(animals['data'])

    return animals
